/*
 * anyRemote
 * a bluetooth remote for your PC.
 * Copyright (C) 2006
 * 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "parse.h"

#define ALARMS_SECTION_STR  "[Alarms]"
#define ALIAS_SECTION_STR   "[Aliases]"
#define KEYS_SECTION_STR    "[Keys]"
#define MODE_END_STR        "[ModeEnd]"
#define MODE_STR    	    "[Mode]"
#define SECTION_END_STR     "[End]"

#define AUTOCONN_STR   	"AutoConnect"
#define AUTOREPEAT_STR	"AutoRepeat"
#define BAUD_STR    	"Baudrate"
#define CHARSET_STR    	"CharSet"
#define CMER_OFF_STR   	"CmerOff"
#define CMER_ON_STR   	"CmerOn"
#define DEBUG_STR   	"Debug"
#define DEVICE_STR   	"Device"
#define LOG_STR    	"Log"
#define RETRY_STR    	"RetrySecs"
#define SCREEN_STR      "Screen"
#define TOMAIN_STR   	"ToMainMenu"
#define MAXTEXT_STR   	"MaxTextSize"

#define EMPTY_STR   	"EMPTY"

#include <regex.h>

#define REGEX_KEYVAL   "^[[:space:]]*(\\[[[:alpha:]]+\\])([^=]*)$|^[[:space:]]*([^=%]*[^[:space:]]{1})[[:space:]]*=[[:space:]]*(.*[^[:space:]]{1})[[:space:]]*$"
#define REGEX_CMDBYCMD "[[:space:]]*(Exec|ExecAndSet|SetMode|Set|Send|ExecAndSend|Timer|CancelTimer|Macro|Load|Include|SendCKPD)[[:space:]]*(\\()|[[:space:]]*(RemoteOn|RemoteOff|Exit|TempDisconnect)[[:space:]]*;{0,1}"

#define REGEX_SET    "[[:space:]]*(bg|editfield|filemanager|fg|font|fullscreen|icons|list|iconlist|menu|parameter|skin|status|text|title|volume|upload)[[:space:]]*,[[:space:]]*(.*)|[[:space:]]*(vibrate|repaint)[[:space:]]*"
#define REGEX_TIMER  "[[:space:]]*([^[:space:]]+)[[:space:]]*,[[:space:]]*([[:digit:]]*)[[:space:]]*,[[:space:]]*([[:digit:]]*)[[:space:]]*,(.*)"

#define REGEX_SET_TL "[[:space:]]*(fg|bg|font|select)[[:space:]]*,(.*)|[[:space:]]*(add|replace)[[:space:]]*,[[:space:]]*([^[:space:]]{1}[^,]*[^[:space:]]{1})[[:space:]]*,[[:space:]]*(.*)|[[:space:]]*(clear|close|show)[[:space:]]*"
#define REGEX_SET_MN "[[:space:]]*(add|replace)[[:space:]]*,(.*)|[[:space:]]*(clear)[[:space:]]*"
#define REGEX_SET_FM "[[:space:]]*(add|replace)[[:space:]]*,[[:space:]]*(left|right)[[:space:]]*,[[:space:]]*(.*)|[[:space:]]*(close|show)[[:space:]]*"

#define SEPARATOR 	'/'

static type_key* findExact(mode *mode, char *key);

regex_t *keyVal   = NULL;
regex_t *cmdByCmd = NULL;

/* Data types */

mode     *modes       = NULL;
mode     *currentMode = NULL;
mode 	 *defMode     = NULL;
type_alias *aliases   = NULL;
type_key *alarms      = NULL;

typedef struct
{
	int   log;
        int   debug;
        int   baudrate;
        int   retrysecs;
        int   screen;
        int   autoConnect;               // AT mode only
        int   autoRepeat;
        char  charset   [MTEXTLEN];     // Charset - AT mode only
        char  device    [MAXARGLEN];    // Device
        char  toMain    [MTEXTLEN];     // AnyMenu  -> MainMenu - AT mode only
        char  cmerOn    [MTEXTLEN];     // AT mode only
        char  cmerOff   [MTEXTLEN];     // AT mode only
        int   model;
        char* cfgDir;
        int   maxText;
        int   frontEnd;
        int   frontEndPort;
} CONF;

CONF conf = { 0, 1, B19200, 60, 0, 0, 0,
                "8859-1",               // charset
                DEFAULT_DEVICE,         // device
		"",   	    	    	// to main menu
    	    	"",     	    	// CMER ON
		"",     	    	// CMER OFF
		MODEL_DEFAULT,
		NULL,
                18,0,5050};                         

/* ----------------- Interface functions ------------------------- */

int getLog()
{
	return conf.log;
}

int getDebug()
{
	return conf.debug;
}

char* getDevice()
{
	return conf.device;
}

int getBaudrate()
{
	return conf.baudrate;
}

int getMaxTextSize()
{
	return conf.maxText;
}

int getRetrySecs()
{
	return conf.retrysecs;
}

int autoConnect()
{
	return conf.autoConnect;
}

int getAutoRepeat()
{
	return conf.autoRepeat;
}

int getUseScreen()
{
	return conf.screen;
}

int getFrontEnd()
{
	return conf.frontEnd;
}

char* getCharset()
{
	return (conf.charset[0] == '\0' ? NULL : conf.charset);
}

char* getToMainMenu()
{
	return  (conf.toMain[0] == '\0' ? NULL : conf.toMain);
}

char* getAT_CMER(int what)
{
	if (what == CMER_ON) {
		if  (conf.cmerOn[0] == '\0') {
			if (conf.model == MODEL_MOTOROLA) {
				return DEF_AT_CMER_ON_MOTOROLA;
			} else if (conf.model == MODEL_SE) {
				return DEF_AT_CMER_ON_SE;
			} else if (conf.model == MODEL_SAGEM) {
				return DEF_AT_CMER_ON_SAGEM;
			} else if (conf.model == MODEL_SIEMENS) {
				return DEF_AT_CMER_ON_SIEMENS;
			} else {
				return DEF_AT_CMER_ON_DEFAULT;
			}
		} else {
			return conf.cmerOn;
		}
	}

	if (what == CMER_OFF) {
		if  (conf.cmerOff[0] == '\0') {
			if (conf.model == MODEL_MOTOROLA) {
				return DEF_AT_CMER_OFF_MOTOROLA;
			} else if (conf.model == MODEL_SE) {
				return DEF_AT_CMER_OFF_SE;
			} else if (conf.model == MODEL_SAGEM) {
				return DEF_AT_CMER_OFF_SAGEM;
			} else if (conf.model == MODEL_SIEMENS) {
				return DEF_AT_CMER_OFF_SIEMENS;
			} else {
				return DEF_AT_CMER_OFF_DEFAULT;
			}
		} else {
			return conf.cmerOff;
		}
	}
	return NULL;
}

char* getAlarmInfo(int which, cmdItem ** cmds)
{
	//logger("DEBUG", "getAlarmInfo()");
	int i;
        type_key *tk = alarms;
        
	for (i=0;i<which;i++) {
        	if (tk != NULL) {
        		tk = tk->next;
                } else {
                	break;
                }
        }
        
        if (i<which || tk == NULL) {
		//logger("DEBUG", "getAlarmInfo() return NULL");
        	return NULL;
        }
    	*cmds = tk->cmd;
    
 	return tk->key;
}

mode* findMode(char *name) 
{
	mode *mp = modes;
	while (mp && strcmp(name, mp->name) != 0) {
		mp = (mode*) mp->next;
	}
	if (mp) {
		sprintf(tmp, "findMode() %s = %s",name,mp->name);
		logger("DEBUG", tmp);
	}
	return mp;
}

void setMode (char *modeName) 
{
	if (modeName ==  NULL) {
               logger("DEBUG", "setMode() NULL input");
               return;
	}
	sprintf(tmp, "setMode() to %s", modeName);
	logger("DEBUG", tmp);

	mode *mp = findMode(modeName);

	if (mp) {
		currentMode = mp;
		sprintf(tmp, "setMode() new mode was set to %s", currentMode->name);
		logger("DEBUG", tmp);
	} else {
		logger("DEBUG", "setMode() new mode did not found");
	}

	return;
}

char *getMode (void) 
{
	return currentMode->name;
}

void setModel (char *answer) 
{
	if (answer == NULL) {
		conf.model = MODEL_DEFAULT;
	} else if (strstr(answer,STR_MOTOROLA)) {
		conf.model = MODEL_MOTOROLA;
	} else if (strstr(answer,STR_SE)) {
		conf.model = MODEL_SE;
	} else if (strstr(answer,STR_SAGEM)) {
		conf.model = MODEL_SAGEM;
	} else if (strstr(answer,STR_SIEMENS)) {
		conf.model = MODEL_SIEMENS;
	} else {
		conf.model = MODEL_DEFAULT;
	}
}

int getModel (void) 
{
	return conf.model;
}

char* getCfgDir(void) 
{
	return conf.cfgDir;
}

void setCfgDir(char *d) 
{
	if (d == NULL) {
        	free(conf.cfgDir);
        } else {
        	conf.cfgDir = strdup(d);
	}
}

/* ----------------- Setup configuration functions ------------------------- */

static int setBaudrate(int rate) 
{
  switch (rate) {
	case 300:
		return B300;
	case 1200:
		return B1200;
	case 2400:
		return B2400;
	case 9600:
		return B9600;
	case 19200:
        	return B19200;
	case 38400:
        	return 38400;
	#ifdef B57600
	case 57600:
		return B57600;
	#endif
	#ifdef B115200
	case 115200:
		return B115200;
	#endif
	#ifdef B230400
	case 230400:
		return B230400;
	#endif
	default:
		printf("WARNING: bad baudrate %d, defaulting to 9600\n", rate);
  }

  return B9600;
}

/* Parses command line options and set flags. */
int parse_opts(int argc, char *argv[])
{
	int i;
	for(i=1; i<argc; i++) {
		if (strcmp(argv[i],"-log")==0 || strcmp(argv[i],"-l")==0) {
			conf.log=1;
		} else if (strcmp(argv[i],"-fe")==0 && i+1<argc) {
			conf.frontEnd=1;
			conf.frontEndPort=atoi(argv[++i]);
		} else if ((strcmp(argv[i],"--serial")==0 || strcmp(argv[i],"-s")==0) && i+1<argc) {
			strcpy(conf.device, argv[++i]);
		} else if ((strcmp(argv[i],"--baud")==0 || strcmp(argv[i],"-b")==0) && i+1<argc) {
			conf.baudrate = setBaudrate(atoi(argv[++i]));
		} else if ((strcmp(argv[i],"--noscreen")==0 || strcmp(argv[i],"-n")==0)) {
			conf.screen = 0;
		} else if ((strcmp(argv[i],"--autoconnect")==0 || strcmp(argv[i],"-a")==0)) {
			conf.autoConnect = 1;
		} else if ((strcmp(argv[i],"-f")==0)) {
			++i;
			continue;   // already processed this parameter
		} else if ((strcmp(argv[i],"-h")==0) || (strcmp(argv[i],"--help")==0)) {
			//Just print help and exit
			printf("Usage: anyremote [-h] [-f conf.file] [-s serial] [-b baudrate] [-n] [-d]\n\n");
			printf(" -h this help\n");
			printf(" -f conf.file: specify non-default conf.file\n");
			printf(" -s serial: specify serial to use\n");
			printf(" -b baudrate: specify baudrate (default is 19200)\n");
			printf(" -n: do not send any AT+CKPD command to the phone\n");
			printf(" -a: reconnect automatically in case of connection failure\n\n");
			printf(" -fe <port>: work as backend for GUI frontend\n\n");
			exit(0);
		} else {
			printf("ERROR: Unknown input parameter %s\n", argv[i]);
		}
	}
	return 1;
}

void regexpPrepare(regex_t *r, char *p) 
{
	int err_no=0;

	if((err_no=regcomp(r, p, REG_EXTENDED))!=0) {
		size_t length; 
		char *buffer;
		length = regerror (err_no, r, NULL, 0);
		buffer = malloc(length);
		regerror (err_no, r, buffer, length);
		printf("regexpPrepare(): regcomp() error %s", buffer);
                
		free(buffer);
		regfree(r);
                
		exit(1);	// exit since we can't parse cfg
	}

	return;
}

void deleteSpaces(char **str, int afterCommas)
{
	char *n = (*str); 
	char *r = (*str);
        int  c  = 0;

	//printf("INFO : deleteSpaces in: >%s<\n", *str);

        while (*r != '\0') { 
		if (isspace(*r) && c >= afterCommas) { 
			r++; 
		} else { 
			if (*r == ',') { 
				c++; 
			} 
			*n = *r;
			r++; 
			n++; 
		} 
	} 
	*n = '\0';

	//printf("INFO : deleteSpaces out: >%s<\n", *str);
}

void normalizeSequence(char **str)
{
	int isSpace = 0;
	
	char *n = (*str); 
	char *r = (*str);

	//printf("INFO : normalizeSequence in: >%s<\n", *str);

	while (*r != '\0') {
		if (isspace(*r)) {
			// Skip spaces
			if(!isSpace) {
				isSpace = 1;
			}
			r++;
		} else if (*r == '(' && r > *str) {  // This is parametrized command Command($$), we have to delete all spaces from brace
			r--;
			deleteSpaces(&r, 0);
			return;
		} else {
			if(isSpace) {
				*n = ' ';
				n++;
			}
			// Copy all symbols
			*n = *r;
			r++;
			n++;
			isSpace = 0;
		}
	}
	*n = '\0';
	    
	//printf("INFO : normalizeSequence out: >%s<\n", *str);
}

regmatch_t* allocRegmatch(int no_sub)
{
	regmatch_t* result;
	if ((result = (regmatch_t *) malloc(sizeof(regmatch_t) * no_sub))==0) {
        	printf("allocRegmatch(): No more memory");
        	exit(1);
        }
        return result;
}

int parseKeyValue(char *in, char ** tag, char ** value)
{
	//printf("INFO: parseKeyValue() %s (%d)\n", in, (int)keyVal->re_nsub+1);

	size_t no_sub = keyVal->re_nsub+1;    
	regmatch_t* result = allocRegmatch(no_sub);

	if (in[strlen(in) - 1] == '\n') {
		in[strlen(in) - 1] = '\0';
	}

	*value =  NULL;
	*tag   =  NULL;
	if (regexec(keyVal, in, no_sub, result, 0)==0) {
		
		int use1 = 1;
		int use2 = 2;
		if (result[1].rm_so == -1) {
			if (result[3].rm_so == -1) {
				printf("parseKeyValue(): Incorrectly formed command (1) %s\n", in);
				use1 = -1;
			}
			use1 = 3;
		} 
		if (result[2].rm_so == -1) {
			if (result[4].rm_so == -1) {
				printf("parseKeyValue(): Incorrectly formed command (2) %s\n", in);
				use2 = -1;
			}
			use2 = 4;
		} 

		if (use1 > 0) {
			*tag   = in + result[use1].rm_so;
			*(in + result[use1].rm_eo) = '\0';
		}
		if (use2 > 0) {
			if (result[use2].rm_so == -1 || result[use2].rm_eo == -1) {
				*value =  NULL;
			} else {  
				*value = in + result[use2].rm_so;
				*(in + result[use2].rm_eo) = '\0';
			}
		}
		/*
		if(*tag != NULL) {
			 printf("INFO : Got tag   : >%s<\n", *tag);
		} else {
			printf("INFO : Got tag : >NULL<\n");
		}
		if(*value != NULL) {
			 printf("INFO : Got value : >%s<\n", *value);
		}else {
			 printf("INFO : Got value : >NULL<\n");
		}
		*/
	//} else {
	//	printf("INFO : Not matched\n");
	}

	free(result);
	return EXIT_OK;
}

static char* id2Cmd (int cmdId)
{
    switch(cmdId) {
        case ID_EXIT:               return CMD_EXIT;
        case ID_TMPDISCONN:         return CMD_TMPDISCONN;
        case ID_REMOTEON:           return CMD_REMOTEON;
        case ID_REMOTEOFF:          return CMD_REMOTEOFF;
        case ID_EXEC:               return CMD_EXEC;
        case ID_SENDCKPD:           return CMD_SENDCKPD;
        case ID_SETMODE:            return CMD_SETMODE;
        case ID_SET:                return CMD_SET;
        case ID_EXECSET:            return CMD_EXECSET;
        case ID_TIMERCMD:           return CMD_TIMER;
        case ID_SEND:               return CMD_SEND;
        case ID_EXECSEND:           return CMD_EXECSEND;
        case ID_CANCELTIMER:        return CMD_CANCELTIMER;
        case ID_MACRO:              return CMD_MACRO;
        case ID_LOAD:               return CMD_LOAD;
        case ID_INCLUDE:            return CMD_INCLUDE;
        
        default:                    return "Unknown";
    }
}

static int cmd2id (char *name) 
{

	if (strcmp(name,CMD_EXIT) == 0) {
        	return ID_EXIT;
        } else if (strcmp(name,CMD_TMPDISCONN) == 0) {
        	return ID_TMPDISCONN;
        } else if (strcmp(name,CMD_REMOTEON) == 0) {
        	return ID_REMOTEON;
        } else if (strcmp(name,CMD_REMOTEOFF) == 0) {
        	return ID_REMOTEOFF;
        } else if (strcmp(name,CMD_EXEC) == 0) {
        	return ID_EXEC;
        } else if (strcmp(name,CMD_SENDCKPD) == 0) {
      		return ID_SENDCKPD;
        } else if (strcmp(name,CMD_SETMODE) == 0) {
		return ID_SETMODE;
        } else if (strcmp(name,CMD_SET) == 0) {
		return ID_SET;
        } else if (strcmp(name,CMD_EXECSET) == 0) {
		return ID_EXECSET;
        } else if (strcmp(name,CMD_TIMER) == 0) {
		return ID_TIMERCMD;
        } else if (strcmp(name,CMD_SEND) == 0) {
		return ID_SEND;
        } else if (strcmp(name,CMD_EXECSEND) == 0) {
		return ID_EXECSEND;
        } else if (strcmp(name,CMD_CANCELTIMER) == 0) {
		return ID_CANCELTIMER;
        } else if (strcmp(name,CMD_MACRO) == 0) {
		return ID_MACRO;
        } else if (strcmp(name,CMD_LOAD) == 0) {
		return ID_LOAD;
        } else if (strcmp(name,CMD_INCLUDE) == 0) {
		return ID_INCLUDE;
        } 
        return ID_UNKNOWN;
}

int storeExecAndSetCmd(cmdItem* ci, char *cmd, int execFlag)
{
	int parseFail = 0;
        
	regex_t* regex  = (regex_t *) malloc(sizeof(regex_t));
	memset(regex, 0, sizeof(regex_t));
                
	regexpPrepare(regex, REGEX_SET);
        
	size_t no_sub = regex->re_nsub+1;	
	
        //printf("storeExecAndSetCmd() %s (%d)\n", cmd, (int)no_sub); 
	
        regmatch_t* result = allocRegmatch(no_sub);
         
	if (regexec(regex, cmd, no_sub, result, 0) == 0) {	// Match it
		
                //printf("storeExecAndSetCmd() matched %d %d\n", result[1].rm_so, result[1].rm_eo);
		
                if (result[1].rm_so >= 0 && result[2].rm_so > 0) {  // subcommand , params matched
                	
                        int l = result[1].rm_eo - result[1].rm_so;
                        
                	if (strncmp(cmd+result[1].rm_so,SET_BG,l)      == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_PARAM,l)   == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_EFIELD,l)  == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_FG,l)      == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_FONT,l)    == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_FSCREEN,l) == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_ICONS,l)   == 0 ||
                       	    strncmp(cmd+result[1].rm_so,SET_SKIN,l)    == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_STATUS,l)  == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_TITLE,l)   == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_UPLD,l)    == 0 ||
                            strncmp(cmd+result[1].rm_so,SET_VOLUME,l)  == 0) {
                            
                                int le = result[2].rm_eo - result[2].rm_so;
				int le2 = 0;
                                
                            	if (!execFlag) {
                                	le2 = le + 1;
                                }
                        	ci->descr = (char*) calloc(1, l + le2 + 1);
                       		strncpy(ci->descr,cmd + result[1].rm_so, l);
                                
                            	if (execFlag) {  
                        		ci->exec = (char*) calloc(1, le + 1);
                       			strncpy(ci->exec,cmd + result[2].rm_so, le);
                                } else {
                                	strcat (ci->descr,",");
                                	strncat(ci->descr,cmd + result[2].rm_so, le);
                                }
                        } else if (strncmp(cmd+result[1].rm_so,SET_FMGR,l) == 0) {
                        
				regex_t* regex2  = (regex_t *) malloc(sizeof(regex_t));
				memset(regex2, 0, sizeof(regex_t));
				regexpPrepare(regex2, REGEX_SET_FM);
				size_t no_sub2 = regex2->re_nsub+1;
        
        			regmatch_t* result2 = allocRegmatch(no_sub2);
                                char* start2 = cmd + result[2].rm_so;
				
                                //printf("parse %s\n", cmd + result[2].rm_so);
                                
                                if (regexec(regex2, cmd + result[2].rm_so, no_sub2, result2, 0) == 0) {
                        
                        
                                	int l2 = 0;
                                
                                	if (result2[1].rm_so >= 0 && result2[2].rm_so >= 0) {  		// add|replace,left|right,_data_
                        
                                        	l2 = result2[1].rm_eo - result2[1].rm_so;
                                                int l3 = result2[2].rm_eo - result2[2].rm_so;
                                                
                                        	int le = result2[3].rm_eo - result2[3].rm_so;
                                        	int le2 = 0;
                                        
                                        	if (!execFlag) {
                                                	le2 = le + 1;
                                                }
                                                
                                        	ci->descr = (char*) calloc(1, l + l2 + l3 + le2 + 3);
                                        	strncpy(ci->descr,cmd + result[1].rm_so, l);

                                        	strcat (ci->descr,",");
                                       		strncat(ci->descr,start2 + result2[1].rm_so, l2);
                                        	strcat (ci->descr,",");
                                       		strncat(ci->descr,start2 + result2[2].rm_so, l3);
                                                
                                                if (execFlag) {
                                        		ci->exec = (char*) calloc(1, le + 1);
                                			strncpy(ci->exec,start2 + result2[3].rm_so, le);
                                               	} else {
                                                	strcat (ci->descr,",");
                                			strncat(ci->descr,start2 + result2[3].rm_so, le);
                                                } 

                        		} else if (result2[4].rm_so >= 0) {				// close|show
                        
                                                l2 = result2[4].rm_eo - result2[4].rm_so;
                                                ci->descr = (char*) calloc(1, l + l2 + 2);
                                                strncpy(ci->descr,cmd + result[1].rm_so, l);
                                        	strcat (ci->descr,",");
                                        	strncat(ci->descr,start2 + result2[4].rm_so, l2);

                                 	} else {
                                                printf("storeExecAndSetCmd(): parse error (F1)\n"); 
                                		parseFail = 1;
                                	}
                                } else {
                                	printf("storeExecAndSetCmd(): parse error (F2)\n"); 
                                	parseFail = 1;
                                }
                         	regfree(regex2);
                                free(result2);
                         
                        } else if (strncmp(cmd+result[1].rm_so,SET_LIST,l) == 0 ||
                                   strncmp(cmd+result[1].rm_so,SET_ILIST,l) == 0 ||
                                   strncmp(cmd+result[1].rm_so,SET_TEXT,l) == 0) {
                                 
				regex_t* regex2  = (regex_t *) malloc(sizeof(regex_t));
				memset(regex2, 0, sizeof(regex_t));
				regexpPrepare(regex2, REGEX_SET_TL);
				size_t no_sub2 = regex2->re_nsub+1;
        
        			regmatch_t* result2 = allocRegmatch(no_sub2);
                                char* start2 = cmd + result[2].rm_so;
				
                                //printf("parse %s\n", cmd + result[2].rm_so);
                                
                                if (regexec(regex2, cmd + result[2].rm_so, no_sub2, result2, 0) == 0) {
                                
                                	int l2 = 0;
                                
                                	if (result2[1].rm_so >= 0) {  		// fg|bg|font|select
                                        
                                        	l2 = result2[1].rm_eo - result2[1].rm_so;
                                        	int le = result2[2].rm_eo - result2[2].rm_so;
                                        	int le2 = 0;
                                        
                                        	if (!execFlag) {
                                                	le2 = le + 1;
                                                }
                                        
                                        	ci->descr = (char*) calloc(1, l + l2 + le2 + 2);
                                        	strncpy(ci->descr,cmd + result[1].rm_so, l);

                                        	strcat (ci->descr,",");

                                       		strncat(ci->descr,start2 + result2[1].rm_so, l2);
                                                
                                                if (execFlag) {
                                        		ci->exec = (char*) calloc(1, le + 1);
                                			strncpy(ci->exec,start2 + result2[2].rm_so, le);
                                               	} else {
                                                	strcat (ci->descr,",");
                                			strncat(ci->descr,start2 + result2[2].rm_so, le);
                                                } 
                                 
                                	} else if (result2[3].rm_so >= 0) {	// add|replace,_title_
                                
                                		l2 = result2[3].rm_eo - result2[3].rm_so;
                                		int l3 = result2[4].rm_eo - result2[4].rm_so;
                                        	int le = result2[5].rm_eo - result2[5].rm_so;
                                        	int le2 = 0;
                                                
                                                if (!execFlag) {
                                                	le2 = le + 1;
                                                }
                                        
                                        	ci->descr = (char*) calloc(1, l + l2 + l3 + le2 + 3);
                                        	strncpy(ci->descr,cmd + result[1].rm_so,  l);
                                        	strcat (ci->descr,",");
                                        	strncat(ci->descr,start2 + result2[3].rm_so, l2);
                                        	strcat (ci->descr,",");
                                        	strncat(ci->descr,start2 + result2[4].rm_so, l3);
                                        	
                                                if (execFlag) {
                                        		ci->exec = (char*) calloc(1, le + 1);
                                			strncpy(ci->exec,start2 + result2[5].rm_so, le);
                                                } else {
                                                	strcat (ci->descr,",");
                                                        strncat(ci->descr,start2 + result2[5].rm_so, le);
                                                }
                                	
                                        } else if (result2[6].rm_so >= 0) {	// close|clear|show
                                        	
                                                l2 = result2[6].rm_eo - result2[6].rm_so;
                                                ci->descr = (char*) calloc(1, l + l2 + 2);
                                                strncpy(ci->descr,cmd + result[1].rm_so, l);
                                        	strcat (ci->descr,",");
                                        	strncat(ci->descr,start2 + result2[6].rm_so, l2);
                                
                                	} else {
                                                printf("storeExecAndSetCmd(): parse error (T1)\n"); 
                                		parseFail = 1;
                                	}
                                } else {
                                	printf("storeExecAndSetCmd(): parse error (T2)\n"); 
                                	parseFail = 1;
                                }
                         	regfree(regex2);
                                free(result2);
                                
                        } else if (strncmp(cmd+result[1].rm_so,SET_MENU,l) == 0) {
                                 
				regex_t* regex2  = (regex_t *) malloc(sizeof(regex_t));
				memset(regex2, 0, sizeof(regex_t));
				regexpPrepare(regex2, REGEX_SET_MN);
				size_t no_sub2 = regex2->re_nsub+1;
        
        			regmatch_t* result2 = allocRegmatch(no_sub2);
                                char* start2 = cmd + result[2].rm_so;
                             	
                                int l2 = 0;
                                
                                if (regexec(regex2, cmd + result[2].rm_so, no_sub2, result2, 0) == 0) {
                                
                                	if (result2[1].rm_so >= 0) {  		// add|replace
                                        
                                        	l2 = result2[1].rm_eo - result2[1].rm_so;
                                        	int le = result2[2].rm_eo - result2[2].rm_so;
                                                int le2 = 0;
                                        	
                                                if (!execFlag) {
                                                	le2 = le + 1;
                                                }
                                                
                                        	ci->descr = (char*) calloc(1, l + l2 + le2 + 2);
                                        	strncpy(ci->descr,cmd + result[1].rm_so, l);

                                        	strcat (ci->descr,",");

                                       		strncat(ci->descr,start2 + result2[1].rm_so, l2);

                                        	if (execFlag) {
                                        		ci->exec = (char*) calloc(1, le + 1);
                                			strncpy(ci->exec,start2 + result2[2].rm_so, le);
                                                } else {
                                        		strcat (ci->descr,",");
                                			strncat(ci->descr,start2 + result2[2].rm_so, le);
                                                 }
                                        
                                        } else if (result2[3].rm_so >= 0) {	// clear

                                                l2 = result2[3].rm_eo - result2[3].rm_so;
                                                ci->descr = (char*) calloc(1, l + l2 + 2);
                                                strncpy(ci->descr,cmd + result[1].rm_so, l);
                                        	strcat (ci->descr,",");
                                        	strncat(ci->descr,start2 + result2[3].rm_so, l2);
                        		
                                        } else {
                                        	printf("storeExecAndSetCmd(): parse error (M1)\n"); 
                                        	parseFail = 1;
                                 	}       
                        	}
                                regfree(regex2);
                                free(result2);
                                
                	} else {
                        	printf("storeExecAndSetCmd(): parse error (M2)\n"); 
                        	parseFail = 1;
        		}
              	} else if (result[3].rm_so >= 0) {   // vibrate,repaint matched
                
                	ci->descr = (char*) calloc(1, result[3].rm_eo - result[3].rm_so + 1);
                        strncpy(ci->descr, cmd + result[3].rm_so, result[3].rm_eo - result[3].rm_so);
		
        	} else {
                	printf("storeExecAndSetCmd(): parse error (1)\n"); 
                	parseFail = 1;
                }
        } else {
        	// ExecAndSet(_shell_commands_) goes here ????
                printf("storeExecAndSetCmd(): parse error (2)\n"); 
        	parseFail = 1;
        }
        
        if (parseFail) {
        	if (execFlag) {
                	printf("storeExecAndSetCmd(): command ExecAndSet( %s ) is formed incorrectly\n", cmd);
                } else { 
        		printf("storeExecAndSetCmd(): command Set( %s ) is formed incorrectly\n", cmd); 
                }
        }
        
        regfree(regex);
	return (parseFail == 1 ? EXIT_NOK : EXIT_OK);
}

int storeTimerCmd(cmdItem* ci, char *cmd)
{
	regex_t* regex  = (regex_t *) malloc(sizeof(regex_t));
	memset(regex, 0, sizeof(regex_t));
	regexpPrepare(regex, REGEX_TIMER);
        
	size_t no_sub = regex->re_nsub+1;	
	
        regmatch_t* result = allocRegmatch(no_sub);
	
        int cmdIsOk = EXIT_OK;
	if (regexec(regex, cmd, no_sub, result, 0) == 0) {	// Match it

                ci->tdata = (timerData*) calloc(1, sizeof(timerData));

                int m = result[1].rm_eo - result[1].rm_so;
                if (m > MAXARGLEN-1) {
                	m = MAXARGLEN-1;
                }
                
                strncpy(ci->tdata->id, cmd + result[1].rm_so, m);

                char iBuff[8];
                int l = result[2].rm_eo - result[2].rm_so;
                if (l > 7) l = 7;
                strncpy(iBuff,cmd + result[2].rm_so,l);
                ci->tdata->timeout = atoi(iBuff);
                
                l = result[3].rm_eo - result[3].rm_so;
                if (l > 7) l = 7;
                strncpy(iBuff,cmd + result[3].rm_so,7);
                ci->tdata->times = atoi(iBuff);
                
                cmdIsOk = storeExecAndSetCmd(ci, cmd + result[4].rm_so, 1);

        	if (cmdIsOk != EXIT_OK) {
                        printf("storeTimerCmd(): command %s is formed incorrectly (1)\n", cmd);
        		free(ci->tdata);
        	}
	} else {
        	printf("storeTimerCmd(): command %s is formed incorrectly (2)\n", cmd);
                cmdIsOk = EXIT_NOK;
        }
        
        regfree(regex);
	return cmdIsOk;
}
         
int storeCmds(cmdItem** cmdList, char *value) 
{
	cmdItem *curCmd = (cmdItem*) NULL;
	int cmdId;

	size_t no_sub = cmdByCmd->re_nsub+1;	
	//printf("storeCmds() %s (%d)\n", value, (int)no_sub); 
	regmatch_t* result = allocRegmatch(no_sub);

	int start = 0;
	while(regexec(cmdByCmd, value+start, no_sub, result, 0)==0) {
		
		//printf("storeCmds() next %s\n", value+start); 

		int  step  = 0;
		char *name = NULL;
		char *cmds = NULL;
	
		int u1 = 1;
		if (result[1].rm_so >= 0 && result[2].rm_so) {  // <name>(<params>) matched  
			cmds = value + start + result[2].rm_so + 1;
			 
			// Search ) which is appropriate to matched (
			int braces = 1; 	// cmd_name( 
			char *ptr = cmds;	  
			while (*ptr != '\0' && braces > 0) {
				if (*ptr == '(') {
					braces++;
				}
				if (*ptr == ')') {
					braces--;
				}
				ptr++;
			} 
			step = ptr - value - start + 1;
			if (*(ptr-1) == ')') {
				*(ptr-1) = '\0';
			}
		} else if (result[3].rm_so >= 0) {  // RemoteOff, RemoteOn, Exit, TempDisconnect matched
			u1 = 3;
			step = result->rm_eo;	     
		} else {
			printf("storeCmds(): Strange match\n");
			start += result->rm_eo;        
			continue;
		}
	
        	// strip spaces from tail
                if (cmds) {
        		char *p = cmds + strlen(cmds) - 1;
        		while (isspace(*p)) {
        			*p = '\0';
                		p--;
        		}  
        	}
                
		name   = value + start + result[u1].rm_so;
		*(value + start + result[u1].rm_eo) = '\0';
	
		if (name == NULL) {
			//printf("Got name >NULL<\n");
			start +=result->rm_eo;        
			continue;
		//} else {
		//	printf("Got name >%s<\n",name);
		}
	
		cmdId = cmd2id(name);

		if (cmdId == ID_UNKNOWN ) {
			printf("storeCmds(): Unknown command name %s\n", name);
			//here could be params like log=... also
			start +=result->rm_eo;        
			continue;
		}
	
		/*
		if (cmds != NULL) {
		    printf("Got command body >%s<\n",cmds);
		} else {
		    printf("Got command body >NULL<\n");
		}
		*/
		//printf("the rest of cmd >%s<\n",value+start+step);
	
        	int cmdIsOk = EXIT_OK;
		// Insert into command list
		cmdItem* newCmd = calloc(1, sizeof(cmdItem));
	
		newCmd->type = cmdId;
		newCmd->next = NULL;
		if (cmds == NULL) {
			newCmd->descr = NULL;
		} else {
			if (cmdId == ID_SET) {
				cmdIsOk = storeExecAndSetCmd(newCmd, cmds, 0);                        	
                        } else if (cmdId == ID_EXECSET) {
				cmdIsOk = storeExecAndSetCmd(newCmd, cmds, 1);                        	
			} else if (cmdId == ID_TIMERCMD) {
				cmdIsOk = storeTimerCmd(newCmd,cmds);
 			} else if (cmdId == ID_EXEC) {
                        	char *newDescr = calloc(1, strlen(cmds) + 1);
				strcpy(newDescr, cmds);
				newCmd->exec = newDescr;
 			} else if (cmdId == ID_EXECSEND) {
                        
                        	char* comma = index(cmds,',');
                                 if (comma == NULL) {
                                	cmdIsOk = EXIT_NOK;
                                } else {
                        
                        		char *dsc = calloc(1, comma - cmds + 1);
                        		char *exc = calloc(1, strlen(comma));
					strncpy(dsc, cmds, comma - cmds);
					strcpy(exc, comma + 1);
					newCmd->descr = dsc;
					newCmd->exec  = exc;
                                }
                        } else {
                        	if (cmdId == ID_SENDCKPD) {
                                	normalizeSequence(&cmds);
                                } else if (cmdId != ID_SEND){
                                	deleteSpaces(&cmds,0);
                                }
                        	char *newDescr = calloc(1, strlen(cmds) + 1);
				strcpy(newDescr, cmds);
				newCmd->descr = newDescr;
                        }
		}
		if (cmdIsOk == EXIT_OK) {
			if (curCmd == NULL) { // Insert first
				*cmdList = newCmd;
			} else {
				curCmd->next = newCmd;
			}
			curCmd = newCmd;
		} else {
                	free(newCmd);
                        if (cmds == NULL) {
                		printf("storeCmds(): command %s was not stored\n", name);
                        } else {
                		printf("storeCmds(): command %s ( %s ) was not stored\n", name,cmds);
                        }
                }
		start +=step;
		//printf("REST OF CMDS %s\n", value+start);	   
	}   // while
	free(result);

	return EXIT_OK;
}

static int storeKey(mode* cMode, type_key **head, char *tag, char *value) 
{
	type_key* It = NULL;
	int ret;

	normalizeSequence(&tag);
        
        if(cMode && findExact(cMode, tag)) {
        	return EXIT_OK;  // do not overwrite existing items
        }

	It = (type_key*) calloc(1, sizeof(type_key));

	// Insert in head
	if ((*head) == NULL) {   // first elem
	 	It->next = NULL;
	} else {
		It->next = (type_key*) (*head);
	}
	*head = It;
	
        It->key = (char*) calloc(1, strlen(tag)+1);
	strcpy(It->key,tag);
	if (value!= NULL) {
		ret = storeCmds(&It->cmd, value);
	} else {
		It->cmd = NULL;
	}

	return EXIT_OK;
}


int loadKeys(FILE *fp, int isInit)
{
	char tmptext[MAXLEN]; 
	char *ret;
	char *tag, *value;    

	if (isInit) {
		// We have at least default mode
		defMode = (mode*) calloc(1, sizeof(mode));
		strcpy(defMode->name,"default");
		defMode->next = NULL;
		defMode->keys = NULL;

		modes	      = defMode;
		currentMode   = defMode;
	} 
       
	mode *curMode = currentMode;
	//int skipFlag = 0;

	do {
		ret=fgets(tmptext, MAXLEN, fp);
	
		if (ret == NULL) return EXIT_NOK;
	
		tag   = NULL;
		value = NULL;
		parseKeyValue(tmptext, &tag, &value);
	
		if (tag == NULL) { 
			//printf("ERROR : incorrect command %s\n", tmptext);
			continue;
		} else if(strcmp(tag,SECTION_END_STR) == 0) {
			break;
		} else if(strcmp(tag,MODE_STR) == 0) { // mode definition
                	mode *nm = findMode(value);
                	if (nm == NULL) {	
                        	//skipFlag = 1;// do not override existing modes
                                //continue;
                
				nm = (mode*) calloc(1, sizeof(mode));
				strcpy(nm->name,value);
				nm->keys = NULL;
			    
				// Insert in head
				if (modes == NULL) {   // first elem
					nm->next = NULL;
				} else {
					nm->next = (mode*) modes;
				}
				modes	= nm;
                        }
			curMode = nm;
	
		} else if(strcmp(tag,MODE_END_STR) == 0) { 	// mode definition end
                	//skipFlag = 0;
			curMode  = defMode;
		} else if (tag != NULL/* && !skipFlag*/) {		// Line with key->command definition
                	//if (isInit || (!isInit && curMode != defMode)) {
				storeKey(curMode, &(curMode->keys), tag, value);
                        //}
		} else {
			continue;
		}
	} while (1);

	return EXIT_OK;    
}

int loadAliases(FILE *fp)
{
	char tmptext[MAXLEN]; 
	char *ret;
	char *key, *alias;
	    

	do {
		ret=fgets(tmptext, MAXLEN, fp);
	
		if (ret == NULL) return EXIT_NOK;
	
		key   = NULL;
		alias = NULL;
		parseKeyValue(tmptext, &key, &alias);
	
		if (key == NULL) { 
			//printf("ERROR : incorrect alias %s\n", tmptext);
			continue;
		} else if(strcmp(key,SECTION_END_STR) == 0) {
			break;
		} else if (alias != NULL) {	// Line with key=alias definition
			
			type_alias* Al = (type_alias*) calloc(1, sizeof(type_alias));

			// Insert in head
			if (aliases == NULL) {   // first elem
				Al->next = NULL;
			} else {
				Al->next = (type_alias*) aliases;
			}
			aliases = Al;

			strcpy(Al->key,  key);
			strcpy(Al->alias,alias);
		} else {
			continue;
		}
	} while (1);

	return EXIT_OK;    
}

int loadAlarms(FILE *fp)
{
	char tmptext[MAXLEN]; 
	char *ret;
	char *tag, *value;    

	do {
		ret=fgets(tmptext, MAXLEN, fp);
	
		if (ret == NULL) return EXIT_NOK;
	
		parseKeyValue(tmptext, &tag, &value);
	
		if (tag == NULL) { 
			continue;
		} else if(strcmp(tag,SECTION_END_STR) == 0) {
			break;
		} else if (value != NULL) {// Line with key->command definition
			storeKey(NULL, &alarms, tag, value);
		} else {
			continue;
		}
	} while (1);

	return EXIT_OK;    
}

int load_cfg(char *mfile, int isInit)
{
	FILE *fp;
	char tmptext[MAXLEN], *ret;
	char *tag, *value;

	fp=fopen(mfile,"r");
	if (fp == NULL) return -1;

	keyVal  = (regex_t *) malloc(sizeof(regex_t));
	memset(keyVal,   0, sizeof(regex_t));
	cmdByCmd = (regex_t *) malloc(sizeof(regex_t));
	memset(cmdByCmd, 0, sizeof(regex_t));

	regexpPrepare(keyVal,   REGEX_KEYVAL);
	regexpPrepare(cmdByCmd, REGEX_CMDBYCMD);

	// Go through the conf file
	while(1) {
		
		ret=fgets(tmptext, MAXLEN, fp);

		if (ret==NULL) break;
		    	
		parseKeyValue(tmptext, &tag, &value);
			
		if (tag == NULL) { 
			//printf("INFO : skip line %s\n", tmptext);
			continue;
		} else if (strcmp(tag,RETRY_STR) == 0 && isInit) {
			if (value != NULL) {
				conf.retrysecs=atoi(value);
			}	    
		} else if (strcmp(tag,DEVICE_STR) == 0 && isInit) {
			if (value != NULL) {
				strcpy(conf.device,value);
			}	    
		} else if (strcmp(tag,AUTOCONN_STR) == 0 && isInit) {
			if (value != NULL && (strncmp("Yes",value,3)==0 || strncmp("true",value,4)==0)) {
			 	conf.autoConnect=1;
			}	    
		} else if (strcmp(tag,CHARSET_STR) == 0 && isInit) {
		    if (value != NULL) {
		    	strcpy(conf.charset,value);
		    }		
		} else if (strcmp(tag,LOG_STR) == 0 && isInit) {
			if (value != NULL && (strncmp("Yes",value,3)==0 || strncmp("true",value,4)==0)) {
				conf.log=1;
			}	    
		} else if (strcmp(tag,SCREEN_STR) == 0 && isInit) {
			if (value != NULL && (strncmp("Yes",value,3)==0 || strncmp("true",value,4)==0)) {
				conf.screen=1;
			}	    
		} else if (strcmp(tag,BAUD_STR) == 0 && isInit) {
			if (value != NULL) {
				conf.baudrate = setBaudrate(atoi(value));
			}	    
		} else if (strcmp(tag,MAXTEXT_STR) == 0 && isInit) {
			if (value != NULL) {
				conf.maxText = atoi(value);
			}	    
		} else if (strcmp(tag,KEYS_SECTION_STR) == 0) {
			if (loadKeys(fp, isInit) != EXIT_OK) {
				return EXIT_NOK;
			} 
                        if (!isInit) {		// If we goes here from Include() command, we did all we need
                        	return EXIT_OK;
                        }
		} else if (strcmp(tag,ALIAS_SECTION_STR) == 0 && isInit) {
			if (loadAliases(fp) != EXIT_OK) {
				return EXIT_NOK;
			}
		} else if (strcmp(tag,ALARMS_SECTION_STR) == 0 && isInit) {
			if (loadAlarms(fp) != EXIT_OK) {
				return EXIT_NOK;
			} 
		} else if (strcmp(tag,TOMAIN_STR) == 0 && isInit) {
			if(value!= NULL) {
				strcpy(conf.toMain,value);
			}	    
		} else if (strcmp(tag,DEBUG_STR) == 0 && isInit) {
			if (value != NULL && (strncmp("No",value,3)==0 || strncmp("false",value,4)==0)) {
				conf.debug=0;
			}	    
		} else if (strcmp(tag,AUTOREPEAT_STR) == 0 && isInit) {
			if (value != NULL && (strncmp("Yes",value,3)==0 || strncmp("true",value,4)==0)) {
				conf.autoRepeat=1;
			}	    
		} else if (strcmp(tag,CMER_ON_STR) == 0 && isInit) {
			if(value!= NULL) {
				strcpy(conf.cmerOn,value);
			}	    
		} else if (strcmp(tag,CMER_OFF_STR) == 0 && isInit) {
			if(value!= NULL) {
				strcpy(conf.cmerOff,value);
			}	    
		} else if (isInit) {
			printf("WARNING: Unknown tag in cfg file >%s<\n",tag);
		}
	}

	fclose(fp);

	return 1;
}

int init_cfg(char *path)
{
	if (path != NULL) {
		
		printf("CFG: Use cfg file >%s<\n",path);
                char *d = strdup(path);
                char *p = rindex(d,SEPARATOR);
                if (p != NULL) {
                 	*p = '\0';
                }
                setCfgDir(d);
                free(d);
                
		return load_cfg(path,1);
	} else {
		char cfgfile[MAXLEN];

		strcpy(cfgfile, getenv("HOME"));
                setCfgDir(cfgfile);
		strcat(cfgfile, CFGFILE);

		printf("CFG: Use cfg file %s\n", cfgfile);
		if(load_cfg(cfgfile,1) != 1) {
			// Try again in current dir 
			strcpy(cfgfile, ".");
                        setCfgDir(cfgfile);
			strcat(cfgfile, CFGFILE);
			
			printf("CFG: Use cfg file %s\n", cfgfile);
			return load_cfg(cfgfile,1);
		}
	}
	return 1;
}

void freeCmds(cmdItem* item)
{
	cmdItem* mi;
        while (item) {
        	mi = item->next;
        	if (item->descr != NULL) {
        		free(item->descr);
        		item->descr = NULL;
        	}
         	if (item->exec != NULL) {
        		free(item->exec);
        		item->exec = NULL;
        	}
         	if (item->tdata != NULL) {
        		free(item->tdata);
        		item->tdata = NULL;
        	}
		
		free(item);
        	item = mi;
        }
	return;
}

void freeCfg(void)
{
	logger("DEBUG", "freeCfg()");
	
        free(conf.cfgDir);

	// Keys
	while(modes) {
		mode	 *md = modes;
		type_key *mk = md->keys;
	
		while (mk) {
			type_key* It = mk;
			mk = (type_key*) mk->next;
			
                        freeCmds(It->cmd);
 
 			It->cmd = NULL;
                        free(It->key);
			free(It);
		}
		modes = (mode*) modes->next;
		free(md);
	}

	// Alarms
        type_key *mk = alarms;
	while (mk) {
		type_key* It = mk;
		mk = (type_key*) mk->next;
		
                freeCmds(It->cmd);

		It->cmd = NULL;
                free(It->key);
		free(It);
	}

	// Aliases
	while(aliases) {
		type_alias *al = aliases->next;
		free(aliases); aliases = NULL;
		aliases = al;
	}

	logger("DEBUG", "freeCfg() OK");
}

/* ------------------------- Utility finctions -------------------------- */

static type_key* findExact(mode *mode, char *key)
{
	//logger("DEBUG", "findExact()");
        
        if (mode == NULL || key == NULL) {
        	logger("DEBUG", "findExact() input is empty ?");
        	return NULL;
        }
	type_key* It = mode->keys;

	// Search exact command
	while (It != NULL && It->key != NULL && strcmp(It->key,key) != 0) {
		//sprintf(tmp,"findExact search >%s< compare to >%s<", It->key, key);
		//logger("DEBUG",tmp);   

		It = (type_key*) It->next;
        	//logger("DEBUG", "findExact() next loop");
	}
        //logger("DEBUG", "findExact() exiting");
	return It;
}

static type_key* findStartingWith(mode *mode, char *key)
{
	//logger("DEBUG", "findStartingWith()");
        if (mode == NULL) {
        	return NULL;
        }
	type_key* It = mode->keys;

	while (It && It->key != NULL) {
		//sprintf(tmp,"findStartingWith search >%s< compare to >%s<", It->key, key);
		//logger("DEBUG",tmp);   
		char *start = strstr(It->key,key);

		if (start && start == It->key) {    // We got a part of multi-key or parametrized command
			
			logger("DEBUG","Found part of multi-key or parametrized command");
			return It;
		} 

		It = (type_key*) It->next;
	}    

	return It;
}


static type_key* findItemInMode(mode *mode, char *key, int *flag, cmdParams *params) 
{
	if (mode == NULL) {
		logger("DEBUG","findItemInMode: mode is null");
		return NULL;
	}

	//sprintf(tmp,"findItemInMode >%s,%s<", mode->name, key);
	//logger("DEBUG",tmp);

	// Prepare to search as parametrized command, contraol presence of both ( and )
	char *start  = index(key,'(');
	char *finish = rindex(key,')');

	if ((start != NULL && finish == NULL) ||  // Command was read partially?
            (start == NULL && finish != NULL)) {  // Command was incorectrly formed ?
		logger("DEBUG","findItemInMode: Incorrectly formed parametrized command. One brace is absent");
		return NULL;
	}
        
	// Clean-up
	if (params != NULL) {
		params->index[0] = '\0';
		params->value[0] = '\0';
	}

	type_key* It = findExact(mode,key);	// Search exact command
	if (It) {
		//logger("DEBUG","Found exact command");
		*flag = FLAG_EXACT;
		return It;
	} 

	It = findStartingWith(mode,key);	// Search as part of multikey sequence
	if (It) {
		//logger("DEBUG","Found part of multi-key command");
		*flag = FLAG_MULTIKEY;
		return It;
	} 

	if (params == NULL) {
		logger("DEBUG","findItemInMode: No parameters suspected. Item does not found.");
		return NULL;
	}

	// Search as parametrized command

	if (start != NULL && finish != NULL && start != key) {        // Do not match "(...)=..."
		It = mode->keys;
	
		char tag  [MAXARGLEN];
		char index[4];
		char value[MAXARGLEN];
	
		strncpy(tag, key, start-key);
		*(tag+(start-key)) = '\0';
	
		char *comma = strstr(start+1,",");
		if (comma == NULL || comma > finish) { // Parametrized command from Java Client should be in a form like List(1,String1)
			logger("DEBUG","findItemInMode: Incorrectly formed parametrized command");
			return NULL;
		}

		strncpy(index,start+1,comma-start-1);
		*(index+(comma-start-1)) = '\0';

		strncpy(value,comma+1,finish-comma-1);
		*(value+(finish-comma-1)) = '\0';

		sprintf(tmp,"Parametrized command parsed as >%s< >%s< >%s< ", tag,index,value);
		logger("DEBUG",tmp);
	
		// Try to search explicitly specified parametrized command (for ex List(1) or List(commandX))
		// By index
		strcat(tag,"(");
		strcat(tag,index);
		strcat(tag,")");
		It = findExact(mode,tag);

		if (It) {
			logger("DEBUG","Found exact (explicitly specified by index) parametrized command");
			*flag = FLAG_EXACT;
			return It;
		} 

		// Then by value
		*(tag+(start-key+1)) = '\0';
		strcat(tag,value);
		strcat(tag,")");
		It = findExact(mode,tag);

		if (It) {
			logger("DEBUG","Found exact (explicitly specified by value) parametrized command");
			*flag = FLAG_EXACT;
			return It;
		} 
	
		// Finally... Search command like ListItem($$)
		*(tag+(start-key)) = '\0';
		
		char* keyAndBrace = calloc(1,strlen(tag)+2);
		strcpy(keyAndBrace,tag);
		strcat(keyAndBrace,"(");
		
                sprintf(tmp,"Search findStartingWith() >%s< ", keyAndBrace);
                logger("DEBUG",tmp);
                
		It = findStartingWith(mode,keyAndBrace);
		free(keyAndBrace);
		if (It) {
			logger("DEBUG","Found parametrized command");

	                sprintf(tmp,"Params >%s,%s< ", index,value);
               		logger("DEBUG",tmp);

			strcpy(params->index, index);
			strcpy(params->value, value);
	
			*flag = FLAG_PARAMETR;
			return It;
		} 
	}

	logger("DEBUG","findItemInMode: Item does not found");
	return NULL;
}

char* findAlias(char *keyIn)
{
	//logger("DEBUG","findAlias");
	type_alias* al = aliases;
	char key[MAXARGLEN];

	if (*keyIn == '\0') {
		//logger("INFO","findAlias: Got empty key");   
		strcpy(key, EMPTY_STR);
	} else {
	 	strncpy(key, keyIn, MAXARGLEN - 1);
	}

	while (al) {
		//sprintf(tmp,"compare alias >%s< = >%s<", key, al->key);
		//logger("DEBUG",tmp);   

		if (strcmp(key, al->key) == 0) {
			sprintf(tmp,"found alias %s=%s", key, al->alias);
			logger("DEBUG",tmp);   
	
			return al->alias;
		}
		al = al->next;
	}
	return NULL;
}

type_key* findItem(char *keyIn, int *flag, cmdParams *params) 
{
	logger("DEBUG","findItem()"); 
	type_key *tk = NULL;

	if (keyIn == NULL) {
        	return NULL;
        }

	char key[MAXARGLEN];

	if (*keyIn == '\0') {
		logger("INFO","Got empty CKEV");   
		strcpy(key, EMPTY_STR);
	} else {
		strncpy(key, keyIn, MAXARGLEN - 1);
	}
	 
	if (currentMode != NULL) {
		tk = findItemInMode(currentMode, key, flag, params);
		if (tk == NULL && currentMode != defMode) {
			logger("DEBUG","findItem() search in default");
			tk = findItemInMode(defMode, key, flag, params);
		}
	}
	return tk;
}

cmdItem *getCommand(type_key* item)
{
	if (!item) {
		logger("DEBUG","getCommand(): Empty type_key was got");
		return NULL;
	}
	if (item->key == NULL) {
		logger("DEBUG","Empty key was got");
		return NULL;
	}

	sprintf(tmp,"getCommand for >%s<", item->key);
	logger("DEBUG",tmp);   

	return item->cmd;
}

//
// Print cfg stuff
//

void printKeysInMode(type_key* It)
{    
        while (It && It->key != NULL) {
        	if (It->cmd) {
    			sprintf(tmp, "\t%s = ", It->key);
        	
        		cmdItem* item = It->cmd;
        		while (item) { 
        			
                                strcat(tmp, id2Cmd(item->type));
                                
                                if (item->tdata || item->descr || item->exec) {
        				strcat(tmp, "(");
                                }
                                
                                if (item->tdata ) {
         				strcat(tmp, item->tdata->id);
                                        
                                        char iBuff[16];
                                        sprintf(iBuff,",%d",item->tdata->timeout);
                               		strcat(tmp, iBuff);
                                        sprintf(iBuff,",%d",item->tdata->times);
                               		strcat(tmp, iBuff);
                                        strcat(tmp, ",");
                                }
                                int pComma = 0;
        			if (item->descr) {
                                	//printf("descr  %s\n",item->descr);
        				strcat(tmp, item->descr);
					pComma = 1;
                                }
        			strcat(tmp, "<|>");
                                if (item->exec) {
                                        if (pComma) {
						strcat(tmp, ",");
        				}
					strcat(tmp, item->exec);
                                }
                                
                                if (item->tdata || item->descr || item->tdata) {
        				strcat(tmp, ")");
        			}
        			strcat(tmp, ";");
        	
        			item = item->next;
        		}
    		} else {
        		sprintf(tmp, "\t%s = no command", It->key);
        	}    
    		logger("CFG", tmp);
	
		It = (type_key*)It->next;
        }
        return;
}

void printKeys()
{    
	mode *mp = modes;
	while (mp) {
		
		sprintf(tmp, "%s %s", MODE_STR, mp->name);
		logger("CFG", tmp);
		 
		type_key* It = mp->keys;
	
		printKeysInMode(It);

		sprintf(tmp, "%s %s", MODE_END_STR, mp->name);
		logger("CFG", tmp);
	
		mp = (mode*)mp->next;
	}

	return;
}

void printAlarms()
{
        logger("CFG", ALARMS_SECTION_STR);

	printKeysInMode(alarms);

        logger("CFG", SECTION_END_STR);
	return;
}

void printAliases()
{
	type_alias *al = aliases;

	logger("CFG", ALIAS_SECTION_STR);

	while(al) {
		sprintf(tmp, "\tKey %s \taliased to %s", al->key, al->alias);
		logger("CFG", tmp);

		al = al->next;
	}
	logger("CFG", SECTION_END_STR);

	return;
}

void printParams()
{
	int atMode = 1;
	if (strstr(conf.device, INET_SOCKET) != NULL || 
	    strstr(conf.device, BT_SOCKET) != NULL) {
		atMode = 0;
	}

	logger("CFG", "[Parameters]");
	if (atMode) {
		sprintf(tmp, "\tAutoConnect   %s", (conf.autoConnect?"Yes":"No"));
		logger("CFG", tmp);
	}
	sprintf(tmp, "\tAutoRepeat    %s", (conf.autoRepeat?"Yes":"No"));
	logger("CFG", tmp);
	if (atMode) {
		if (conf.cmerOn[0] != '\0') {
			sprintf(tmp, "\tCMER On       %s", conf.cmerOn);
			logger("CFG", tmp);
		} else {
			logger("CFG", "\tCMER On	   Will be determined automatically");
		}
		if (conf.cmerOff[0] != '\0') {
			sprintf(tmp, "\tCMER Off      %s", conf.cmerOff);
			logger("CFG", tmp);
		} else {
			logger("CFG", "\tCMER Off	   Will be determined automatically");  
		}
	}
	sprintf(tmp, "\tDebug	      %s", (conf.debug?"Yes":"No"));
	logger("CFG", tmp);
	sprintf(tmp, "\tDevice        %s", conf.device);
	logger("CFG", tmp);
	sprintf(tmp, "\tLog	      %s", (conf.log?"Yes":"No"));
	logger("CFG", tmp);
	if (!atMode) {
		sprintf(tmp, "\tMax Text Size %d", conf.maxText);
		logger("CFG", tmp);
	}
	sprintf(tmp, "\tRetry Sec     %d", conf.retrysecs);
	logger("CFG", tmp);
	sprintf(tmp, "\tScreen Menu   %s", (conf.screen?"Yes":"No"));
	logger("CFG", tmp);
	if (atMode) {
		sprintf(tmp, "\tTo Main Menu  %s", conf.toMain);
		logger("CFG", tmp);
	}
	if (conf.frontEnd) {
		sprintf(tmp, "\tWorks as back end");
		sprintf(tmp, "\tPort          %d", conf.frontEndPort);
		logger("CFG", tmp);
	}
	logger("CFG", SECTION_END_STR);
} 

void printConf() 
{
	printKeys   ();
	printAliases();
	printAlarms ();
	printParams ();

	return ;
}
